//---------------------------------
// 2018/7/8

#include "WebPageHandler.h"
#include "Storage.h"
#include "TimelyTaskManager.h"
#include "INA219I2C.h"
#include "StrList.h"

extern INA219I2C ina219;
extern TimelyTaskManager timelyTaskManager;
extern bool logMode;
extern bool load;
extern bool loadCut;
extern bool sleepMode;
extern String logDir;
extern AsyncWebServer server;

extern float vBatTarget;
extern float cBatTarget;
extern float vLoadCut;
extern float vLoadRecover;

// Prototype of SolarCharger.ino functions
String vchargeStr();
void startLog();
void stopLog();
void loadControl();
void sleepControl();

// Web server file path
const char WebPageHandler::pageBattery[] = "/battery.php";
const char WebPageHandler::pageLogFile[] = "/logfile.php";
const char WebPageHandler::pageSystem[] = "/system.php";

// Callback
std::function<void(uint8_t)> WebPageHandler::callBack;


// Initialize and start web server
void WebPageHandler::startServer(std::function<void(uint8_t)> f){
  callBack = f;

  server.on(pageBattery, HTTP_ANY, handleBatteryPage);
  server.on(pageLogFile, HTTP_ANY, handleLogFilePage);
  server.on(pageSystem, HTTP_ANY, handleSystemPage);
  server.on((Storage::getOpLogPath()).c_str(), HTTP_ANY, handleDownload);
  if(Storage::opLogSD){
    server.serveStatic("/", SD, "/");
  }else{
    server.serveStatic("/", SPIFFS, "/");
  }
  server.onNotFound(handleNotFound);
  server.begin();
}

// Print HTML header
void WebPageHandler::printHeader(AsyncResponseStream *response){
  response->print(
    "<head>\n"
    "  <meta charset=\"utf-8\">\n"
    "  <meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n"
    "  <meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">\n"
    "  <title>E32-SolarCharger</title>\n\n"
    
    "  <!-- Bootswatch flatly -->\n"
    "  <link href=\"https://maxcdn.bootstrapcdn.com/bootswatch/3.3.7/flatly/bootstrap.min.css\" rel=\"stylesheet\" integrity=\"sha384-+ENW/yibaokMnme+vBLnHMphUYxHs34h9lpdbSLuAwGkOKFRl4C34WkjazBtb7eT\" crossorigin=\"anonymous\">\n"
    "</head>\n"
  );
}

// Print HTML navigation bar
void WebPageHandler::printNavBar(AsyncResponseStream *response, const char*page){
  response->print(
    "<nav class=\"navbar navbar-default navbar-fixed-top\">\n"
    "  <div class=\"container-fluid\">\n"
    "    <div class=\"navbar-header\">\n"
    "      <button type=\"button\" class=\"navbar-toggle collapsed\" data-toggle=\"collapse\" data-target=\"#bs-example-navbar-collapse-1\">\n"
    "        <span class=\"sr-only\">Toggle navigation</span>\n"
    "        <span class=\"icon-bar\"></span>\n"
    "        <span class=\"icon-bar\"></span>\n"
    "        <span class=\"icon-bar\"></span>\n"
    "      </button>\n"
    "      <a class=\"navbar-brand\" href=\"#\">E32-SolarCharger</a>\n"
    "    </div>\n\n"
  );
  response->print(
    "    <div class=\"collapse navbar-collapse\" id=\"bs-example-navbar-collapse-1\">\n"
    "      <ul class=\"nav navbar-nav\">\n"
  );
  
  if(0==strcmp(page, pageBattery)){
    response->print("        <li class=\"active\"><a href=\"" + String(pageBattery) + "\">Battery<span class=\"sr-only\">(current)</span></a></li>\n");
  }else{
    response->print("        <li><a href=\"" + String(pageBattery) + "\">Battery</a></li>\n");
  }
  if(0==strcmp(page, pageLogFile)){
    response->print("        <li class=\"active\"><a href=\"" + String(pageLogFile) + "\">Log File<span class=\"sr-only\">(current)</span></a></li>\n");
  }else{
    response->print("        <li><a href=\"" + String(pageLogFile) + "\">Log File</a></li>\n");
  }
  if(0==strcmp(page, pageSystem)){
    response->print("        <li class=\"active\"><a href=\"" + String(pageSystem) + "\">System<span class=\"sr-only\">(current)</span></a></li>\n");
  }else{
    response->print("        <li><a href=\"" + String(pageSystem) + "\">System</a></li>\n");
  }
  response->print(
    "      </ul>\n"
    "    </div>\n"
    "  </div>\n"
    "</nav>\n\n"
  );
}

// Print bootstrap script
void WebPageHandler::printScript(AsyncResponseStream *response){
  response->print(
    "<script src=\"https://ajax.googleapis.com/ajax/libs/jquery/1.12.4/jquery.min.js\"></script>\n"
    "<script src=\"https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/js/bootstrap.min.js\" integrity=\"sha384-Tc5IQib027qvyjSMfHjOMaLkfuWVxZxUPnCJA7l2mCWNIpG9mGCD8wGNIcPD7Txa\" crossorigin=\"anonymous\"></script>\n"
  );
}


void WebPageHandler::handleBatteryPage(AsyncWebServerRequest *request){
  String actionStr;
  if(request->hasArg("action")){
    actionStr = request->arg("action");
  }
  if(actionStr == "startLog"){
    startLog();
  }else if(actionStr == "stopLog"){
    stopLog();
  }else if(actionStr == "sleepAuto"){
    sleepMode = true; // Sleep control() is after responding HTTP request
  }else if(actionStr == "sleepOff"){
    sleepMode = false;
  }else if(actionStr == "loadAuto"){
    load = true;
    loadControl();
    delay(100);
  }else if(actionStr == "loadOff"){
    load = false;
    loadControl();
    delay(100);
  }
  
  AsyncResponseStream *response = request->beginResponseStream("text/html");
  
  response->print(
    "<!DOCTYPE html>\n\n"
    "<html lang=\"ja\">\n\n"
  );
  printHeader(response);
  response->print(
    "<body style=\"padding-top:70px\">\n\n"
  );
  printNavBar(response, pageBattery);
  
  response->print(
    "<div class=\"container\">\n"
    "  <div class=\"row\"><div class=\"col-sm-10 col-md-8\">\n"
    "    <div class=\"panel panel-primary\">\n"
    "      <div class=\"panel-heading\">\n"
    "        <h3 class=\"panel-title\">Battery Status</h3>\n"
    "      </div>\n"
    "      <div class=\"panel-body\">\n"
  );
  char p[8];
  response->print(String()+
    "        <h4>Max Charge Voltage : " + dtostrf(vBatTarget, 1, 1, p) + "V</h4>\n"
    "        <h4>Max Charge Current : " + dtostrf(cBatTarget, 1, 1, p) + "A</h4>\n"
    "        <h4>Load Cut Voltage : " + dtostrf(vLoadCut, 1, 1, p) + "V</h4>\n"
    "        <h4>Load Recovery Voltage : " + dtostrf(vLoadRecover, 1, 1, p) + "V</h4>\n"
    "        <h4>Battery Voltage : " + ina219.vbusStr(false) + "V</h4>\n"
    "        <h4>Battery Current : " + ina219.currentStr(false) + "A</h4>\n"
    "        <h4>Battery Power : " + dtostrf(ina219.vbus(false) * ina219.current(false), 1, 2, p) + "W</h4>\n"
    "        <h4>Input Voltage : " + vchargeStr() + "V</h4>\n"
  );
  
  if(logMode){
    response->print("        <h4>Log : Running\n");
  }else{
    response->print("        <h4>Log : Stop\n");
  }
  if(sleepMode){
    response->print("        <h4>Sleep : Auto\n");
  }else{
    response->print("        <h4>Sleep : Off\n");
  }
  if(load){
    response->print("        <h4>Load : Auto\n");
  }else{
    response->print("        <h4>Load : Off\n");
  }
  response->print(
    "      </div>\n"
    "    </div>\n"
  );
  response->print(
    "    <form method=\"POST\" class=\"form-horizontal\">\n"
    "      <div class=\"form-group\">\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"startLog\" class=\"btn btn-default btn-block\">Start Recording</button>\n"
    "        </div>\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"stopLog\" class=\"btn btn-default btn-block\">Stop Recording</button>\n"
    "        </div>\n"
    "      </div>\n"
    "      <div class=\"form-group\">\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"sleepAuto\" class=\"btn btn-default btn-block\">Sleep Auto</button>\n"
    "        </div>\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"sleepOff\" class=\"btn btn-default btn-block\">Sleep OFF</button>\n"
    "        </div>\n"
    "      </div>\n"
    "      <div class=\"form-group\">\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"loadAuto\" class=\"btn btn-default btn-block\">Load Auto</button>\n"
    "        </div>\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"loadOff\" class=\"btn btn-default btn-block\">Load OFF</button>\n"
    "        </div>\n"
    "      </div>\n"
    "    </form>\n"
  );
  response->print(  
    "  </div></div><!-- row, col -->\n"
    "</div>\n\n"
  );

  printScript(response);

  response->print(
    "</body>\n"
    "</html>\n"
  );

  request->send(response);

  if(actionStr == "sleepAuto"){
    timelyTaskManager.setOneTimeTask(idSleepControlOneTime, sleepControl);
  }
}


void WebPageHandler::handleLogFilePage(AsyncWebServerRequest *request){
  String selectFile = "";
  String filelistStr;
  String actionStr;
  
  if(request->hasArg("filelist")){
    filelistStr = request->arg("filelist");
  }
  if(request->hasArg("action")){
    actionStr = request->arg("action");
  }
  
  if(actionStr == "download"){
    Storage::opLog(timelyTaskManager.getTimeStr() + " Download Log File " + filelistStr + "\n");
    String path = logDir + "/" + filelistStr;
    File file = Storage::open(path, FILE_READ);
    if(file){
      file.close();
      if(Storage::opLogSD){
        request->send(SD, path, "application/octet-stream");
      }else{
        request->send(SPIFFS, path, "application/octet-stream");
      }
    }
    return;
  }else if(actionStr == "delete"){
    Storage::opLog(timelyTaskManager.getTimeStr() + " Delete Log File " + filelistStr + "\n");
    String path = logDir + "/" + filelistStr;
    Storage::remove(path);
    Storage::end();
  }else if(actionStr == "deleteall"){
    Storage::opLog(timelyTaskManager.getTimeStr() + " Delete All Log Files " + filelistStr + "\n");
    File root = Storage::open(logDir, FILE_READ);
    if(root){
      if(root.isDirectory()){
        File file = root.openNextFile();
        while(file){
          String path = file.name();
          file.close();
          Storage::remove(path);
          file = root.openNextFile();
        }
      }
    }
    Storage::end();
  }

  AsyncResponseStream *response = request->beginResponseStream("text/html");
  
  response->print(
    "<!DOCTYPE html>\n\n"
    "<html lang=\"ja\">\n\n"
  );
  printHeader(response);
  response->print(
    "<body style=\"padding-top:70px\">\n\n"
  );
  printNavBar(response, pageLogFile);
  
  response->print(
    "<div class=\"container\">\n"
    "  <div class=\"row\"><div class=\"col-sm-10 col-md-8\">\n"
  );
  
  response->print(
    "    <form method=\"POST\" class=\"form-horizontal\">\n"
    "      <div class=\"form-group\"><div class=\"col-xs-12\">\n"
    "        <select class=\"form-control\" size=\"8\" name=\"filelist\">\n"
  );
  
  // Read directory
  uint16_t fileCount = 0;
  StrList*strList = new StrList();
  File root = Storage::open(logDir, FILE_READ);
  if(root){
    if(root.isDirectory()){
      File file = root.openNextFile();
      while(file){
        String fname = file.name();
        fname.replace(logDir + "/", "");
        strList->add(fname);
        fileCount++;
        file = root.openNextFile();
      }
    }
  }
  Storage::end();

  StrListItem*strListItem = strList->getFirst();

  for(uint16_t i=0; i<fileCount; i++){
    response->print("          <option>" + strListItem->str + "</option>\n");
    strListItem = strList->getNext();
  }
  delete strList;
  
  response->print(
    "        </select>\n"
    "      </div></div>\n"
    "      <div class=\"form-group\">\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"download\" class=\"btn btn-default btn-block\">Download</button>\n"
    "        </div>\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"submit\" name=\"action\" value=\"delete\" class=\"btn btn-default btn-block\">Delete</button>\n"
    "        </div>\n"
    "      </div>\n"
    "      <div class=\"form-group\">\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"button\" class=\"btn btn-default btn-block\" data-toggle=\"modal\" data-target=\"#deleteAllDialog\">Delete All</button>\n"
    "        </div>\n"
    "      </div>\n"
    "    </form>\n"
  );

  response->print(
    "  </div></div><!-- row, col -->\n"
    "</div>\n"
  );
  
  response->print(
    "<div class=\"modal\" id=\"deleteAllDialog\">\n"
    "  <div class=\"modal-dialog\">\n"
    "    <div class=\"modal-content\">\n"
    "      <div class=\"modal-header\">\n"
    "        <button type=\"button\" class=\"close\" data-dismiss=\"modal\" aria-hidden=\"true\">&times;</button>\n"
    "        <h4 class=\"modal-title\">Delete All Files?</h4>\n"
    "      </div>\n"
    "      <!-- <div class=\"modal-body\">\n"
    "      </div> -->\n"
    "      <div class=\"modal-footer\">\n"
    "        <form method=\"POST\">\n"
    "          <button type=\"button\" class=\"btn btn-default\" data-dismiss=\"modal\">Cancel</button>\n"
    "          <button type=\"submit\" class=\"btn btn-primary\" name=\"action\" value=\"deleteall\">OK</button>\n"
    "        </form>\n"
    "      </div>\n"
    "    </div>\n"
    "  </div>\n"
    "</div>\n\n"
  );

  printScript(response);

  response->print(
    "</body>\n"
    "</html>\n"
  );

  request->send(response);
}



void WebPageHandler::handleSystemPage(AsyncWebServerRequest *request){
  String actionStr;;
  if(request->hasArg("action")){
    actionStr = request->arg("action");
  }
  
  if(actionStr == "formattorage"){
    if(!Storage::opLogSD){
      SPIFFS.format();
    }
    Storage::opLog(timelyTaskManager.getTimeStr() + " Format Storage\n");
  }
  
  if(actionStr == "clearOpLog"){
    String path = Storage::getOpLogPath();
    Storage::remove(path);
    Storage::end();
    Storage::opLog(timelyTaskManager.getTimeStr() + " Clear Op Log\n");
  }
  
  AsyncResponseStream *response = request->beginResponseStream("text/html");
  
  response->print(
    "<!DOCTYPE html>\n\n"
    "<html lang=\"ja\">\n\n"
  );
  printHeader(response);
  response->print(
    "<body style=\"padding-top:70px\">\n\n"
  );
  printNavBar(response, pageSystem);
  
  response->print(
    "<div class=\"container\">\n"
    "  <div class=\"row\"><div class=\"col-sm-10 col-md-8\">\n"
    "    <div class=\"panel panel-primary\">\n"
    "      <div class=\"panel-heading\">\n"
    "        <h3 class=\"panel-title\">System Status</h3>\n"
    "      </div>\n"
    "      <div class=\"panel-body\">\n"
  );
  if(Storage::opLogSD){
    response->print("        <h4>Storage : SD</h4>\n");
  }else{
    SPIFFS.begin(true);
    response->print("        <h4>Storage : SPIFFS</h4>\n");
    response->print("        <h5>Total : " + String(SPIFFS.totalBytes()) + " Bytes</h5>\n");
    response->print("        <h5>Used : " + String(SPIFFS.usedBytes()) + " Bytes</h5>\n");
  }
  response->print("        <h4>Flash Size : " + String(ESP.getFlashChipSize()) + " Bytes</h4>\n");
  response->print("        <h4>Flash Frequency : " + String(ESP.getFlashChipSpeed()/1000000) + " MHz</h4>\n");
  response->print("        <h4>Free Heap : " + String(ESP.getFreeHeap()) + " Bytes</h4>\n");
  if(Storage::opLogSerial){
    response->print("        <h4>Operation Log Serial : Enable</h4>\n");
  }else{
    response->print("        <h4>Operation Log Serial : Disable</h4>\n");
  }
  if(Storage::opLogStorage){
    response->print("        <h4>Operation Log Storage : Enable</h4>\n");
  }else{
    response->print("        <h4>Operation Log Storage : Disable</h4>\n");
  }
  
  response->print(
    "      </div>\n"
    "    </div>\n"
    "    <form methos=\"POST\" class=\"form-horizontal\"><fieldset>\n"
    "      <div class=\"form-group\">\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"button\" class=\"btn btn-default btn-block\" data-toggle=\"modal\" data-target=\"#clearOpLogDialog\">Clear Op Log</button>\n"
    "        </div>\n"
    "        <div class=\"col-xs-6\">\n"
    "          <button type=\"button\" class=\"btn btn-default btn-block\" onclick=\"location.href='"
  );
  response->print(Storage::getOpLogPath());
  
  response->print(
    "'\">Download Op Log</button>\n"
    "        </div>\n"
    "      </div>\n"
  );
  if(!Storage::opLogSD){
    response->print(
      "      <div class=\"form-group\">\n"
      "        <div class=\"col-xs-6\">\n"
      "          <button type=\"button\" class=\"btn btn-default btn-block\" data-toggle=\"modal\" data-target=\"#formattorageDialog\">Format Storage</button>\n"
      "        </div>\n"
      "      </div>\n"
    );
  }
  response->print(
    "    </fieldset></form>\n"
    "  </div></div><!-- row, col -->\n"
    "</div>\n\n"
    "<div class=\"modal\" id=\"clearOpLogDialog\">\n"
    "  <div class=\"modal-dialog\">\n"
    "    <div class=\"modal-content\">\n"
    "      <div class=\"modal-header\">\n"
    "        <button type=\"button\" class=\"close\" data-dismiss=\"modal\" aria-hidden=\"true\">&times;</button>\n"
    "        <h4 class=\"modal-title\">Clear Op Log file?</h4>\n"
    "      </div>\n"
    "      <!-- <div class=\"modal-body\">\n"
    "      </div> -->\n"
    "      <div class=\"modal-footer\">\n"
    "        <form methos=\"POST\">\n"
    "          <button type=\"button\" class=\"btn btn-default\" data-dismiss=\"modal\">Cancel</button>\n"
    "          <button type=\"submit\" class=\"btn btn-primary\" name=\"action\" value=\"clearOpLog\">OK</button>\n"
    "        </form>\n"
    "      </div>\n"
    "    </div>\n"
    "  </div>\n"
    "</div>\n\n"
    "<div class=\"modal\" id=\"formattorageDialog\">\n"
    "  <div class=\"modal-dialog\">\n"
    "    <div class=\"modal-content\">\n"
    "      <div class=\"modal-header\">\n"
    "        <button type=\"button\" class=\"close\" data-dismiss=\"modal\" aria-hidden=\"true\">&times;</button>\n"
    "        <h4 class=\"modal-title\">Format Storage?</h4>\n"
    "      </div>\n"
    "      <!-- <div class=\"modal-body\">\n"
    "      </div> -->\n"
    "      <div class=\"modal-footer\">\n"
    "        <form methos=\"POST\">\n"
    "          <button type=\"button\" class=\"btn btn-default\" data-dismiss=\"modal\">Cancel</button>\n"
    "          <button type=\"submit\" class=\"btn btn-primary\" name=\"action\" value=\"formattorage\">OK</button>\n"
    "        </form>\n"
    "      </div>\n"
    "    </div>\n"
    "  </div>\n"
    "</div>\n\n"
  );

  printScript(response);

  response->print(
    "</body>\n"
    "</html>\n"
  );

  request->send(response);
}

void WebPageHandler::handleDownload(AsyncWebServerRequest *request){
  Storage::opLog(timelyTaskManager.getTimeStr() + " Download " + request->url() + "\n");
  File file = Storage::open(request->url(), FILE_READ);
  if(file){
    file.close();
    if(Storage::opLogSD){
      request->send(SD, request->url(), "application/octet-stream");
    }else{
      request->send(SPIFFS, request->url(), "application/octet-stream");
    }
  }else{
    request->send(404);
    Storage::end();
  }
}

void WebPageHandler::handleNotFound(AsyncWebServerRequest *request){
  request->send(404);
}




